__version__ = "0.1"
__all__ = []

from pyiron.project import Project
from pyiron.atomistics.structure.atoms import ase_to_pyiron, pyiron_to_ase, Atoms
from pyiron.base.job.script import Notebook
from pyiron.base.settings.install import install_dialog

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions


def install():
    install_dialog()
